"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  This file is part of the Smart Developer Hub Project:
    http://www.smartdeveloperhub.org

  Center for Open Middleware
        http://www.centeropenmiddleware.com/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2015 Center for Open Middleware.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at 

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
import logging
from random import random

from abc import ABCMeta, abstractmethod
from agora.client.wrapper import Agora
from agora.stoa.actions.core import STOA, RDF, AGENT_ID
from agora.stoa.actions.core.delivery import DeliveryRequest, DeliveryAction, DeliveryResponse, DeliverySink
from agora.stoa.actions.core.utils import CGraph, GraphPattern
from agora.stoa.actions.core.utils import tp_parts
from agora.stoa.server import app
from agora.stoa.store import r
from rdflib import Literal, URIRef
from shortuuid import uuid

__author__ = 'Fernando Serena'

log = logging.getLogger('agora.stoa.actions.fragment')
agora_conf = app.config['AGORA']
PASS_THRESHOLD = app.config['BEHAVIOUR']['pass_threshold']
agora_client = Agora(**agora_conf)
fragment_locks = {}

# Ping Agora
try:
    _ = agora_client.prefixes
except Exception:
    log.warning('Agora is not currently available at {}'.format(agora_conf))
else:
    log.info('Connected to Agora: {}'.format(agora_conf))

log.info("""Behaviour parameters:
                    - Pass threshold: {}""".format(PASS_THRESHOLD))


class FragmentRequest(DeliveryRequest):
    def __init__(self):
        """
        Prepares the pattern graph object which will store the one that is contained in the
        request
        """
        super(FragmentRequest, self).__init__()
        self.__request_graph = CGraph()
        self.__request_graph.bind('stoa', STOA)
        self.__preferred_labels = set([])
        self.__variable_labels = {}
        self._graph_pattern = GraphPattern()

        # Copy Agora prefixes to the pattern graph
        try:
            prefixes = agora_client.prefixes
            for p in prefixes:
                self.__request_graph.bind(p, prefixes[p])
        except Exception, e:
            raise EnvironmentError(e.message)

    def _extract_content(self, request_type=None):
        super(FragmentRequest, self)._extract_content(request_type)

        # Firstly, it looks for all Variables that are inside the request
        variables = set(self._graph.subjects(RDF.type, STOA.Variable))
        if not variables:
            raise SyntaxError('There are no variables specified for this request')
        log.debug(
            'Found {} variables in the the fragment pattern'.format(len(variables)))

        # Secondly, try to identify all links between variables (avoiding cycles)
        visited = set([])
        for v in variables:
            self.__request_graph.add((v, RDF.type, STOA.Variable))
            self.__follow_variable(v, visited=visited)

        # Thirdly, the request graph is filtered and the request pattern only contains
        # the relevant nodes and their relations

        # Finally, an Agora-compliant Graph Pattern is created and offered as a property
        self.__build_graph_pattern()

        log.debug('Extracted (fragment) pattern graph:\n{}'.format(self.__request_graph.serialize(format='turtle')))

    def __n3(self, elm):
        """
        :param elm: The element to be n3-formatted
        :return: The n3 representation of elm
        """
        return elm.n3(self.__request_graph.namespace_manager)

    def __follow_variable(self, variable_node, visited=None):
        """
        Recursively follows one variable node of the request graph
        :param variable_node: Starting node
        :param visited: Track of visited variable nodes
        :return:
        """

        def add_pattern_link(node, triple):
            type_triple = (node, RDF.type, STOA.Variable)
            condition = type_triple in self._graph
            if condition:
                self.__request_graph.add(type_triple)
            elif isinstance(node, URIRef):
                condition = True
            if condition:
                if triple not in self.__request_graph:
                    self.__request_graph.add(triple)
                    log.debug('New pattern link: {}'.format(triple))
            return condition

        if visited is None:
            visited = set([])
        visited.add(variable_node)
        subject_pattern = self._graph.subject_predicates(variable_node)
        for (n, pr) in subject_pattern:
            if add_pattern_link(n, (n, pr, variable_node)) and n not in visited:
                self.__follow_variable(n, visited)

        object_pattern = self._graph.predicate_objects(variable_node)
        for (pr, n) in object_pattern:
            if add_pattern_link(n, (variable_node, pr, n)):
                if n not in visited:
                    self.__follow_variable(n, visited)
            elif n != STOA.Variable:
                self.__request_graph.add((variable_node, pr, n))

    def __build_graph_pattern(self):
        """
        Creates a GraphPattern with all the identified (Agora compliant) triple patterns
        in the request graph
        """

        def preferred_label():
            # Each variable may have a property STOA.label that specifies its desired label
            labels = list(self.__request_graph.objects(v, STOA.label))
            p_label = labels.pop() if len(labels) == 1 else ''
            if p_label:
                self.__preferred_labels.add(str(p_label))
            return p_label if p_label.startswith('?') else '?v{}'.format(i)

        # Populates a dictionary with all variables and their labels
        variables = self.__request_graph.subjects(RDF.type, STOA.Variable)
        for i, v in enumerate(variables):
            self.__variable_labels[v] = preferred_label()

        # For each variable, generates one triple pattern per relation with other nodes as either subject or object
        for v in self.__variable_labels.keys():
            v_in = self.__request_graph.subject_predicates(v)
            for (s, pr) in v_in:
                s_part = self.__variable_labels[s] if s in self.__variable_labels else self.__n3(s)
                self._graph_pattern.add(u'{} {} {}'.format(s_part, self.__n3(pr), self.__variable_labels[v]))
            v_out = self.__request_graph.predicate_objects(v)
            for (pr, o) in [_ for _ in v_out if _[1] != STOA.Variable and not _[0].startswith(STOA)]:
                o_part = self.__variable_labels[o] if o in self.__variable_labels else (
                    '"{}"'.format(o) if isinstance(o, Literal) else self.__n3(o))
                p_part = self.__n3(pr) if pr != RDF.type else 'a'
                self._graph_pattern.add(u'{} {} {}'.format(self.__variable_labels[v], p_part, o_part))

    @property
    def pattern(self):
        """
        :return: The request graph pattern
        """
        return self._graph_pattern

    @property
    def preferred_labels(self):
        """
        :return: The variable preferred labels
        """
        return self.__preferred_labels

    @property
    def variable_labels(self):
        """
        :return: All variable labels
        """
        return self.__variable_labels.values()

    def variable_label(self, n):
        label = self.__variable_labels.get(n, None)
        if isinstance(label, Literal):
            label = label.toPython()
        return label


class FragmentAction(DeliveryAction):
    __metaclass__ = ABCMeta

    def __init__(self, message):
        super(FragmentAction, self).__init__(message)


class FragmentSink(DeliverySink):
    __metaclass__ = ABCMeta

    def __init__(self):
        super(FragmentSink, self).__init__()
        self._graph_pattern = GraphPattern()
        self._fragment_pattern = GraphPattern()
        self._filter_mapping = {}
        self._fragments_key = '{}:fragments'.format(AGENT_ID)
        self.__f_key_pattern = '{}:'.format(self._fragments_key) + '{}'
        self._fragment_key = None
        self._preferred_labels = set([])

    def __check_gp_mappings(self, gp=None):
        """
        Used in _save method. Seeks matches with some fragment already registered
        :param gp: By default, _graph_pattern attribute is used when gp is None
        :return: The matching fragment id and the mapping dictionary or None if there is no matching
        """
        if gp is None:
            gp = self._graph_pattern
        gp_keys = r.keys('{}:*:gp'.format(self._fragments_key))
        for gpk in gp_keys:
            stored_gp = GraphPattern(r.smembers(gpk))
            mapping = stored_gp.mapping(gp)
            if mapping:
                return gpk.split(':')[-2], mapping
        return None

    def _remove_tp_filters(self, tp):
        """
        Transforms a triple pattern that may contain filters to a new one with both subject and object bounded
        to variables
        :param tp: The triple pattern to be filtered
        :return: Filtered triple pattern
        """

        def __create_var(elm, predicate):
            if elm in self._filter_mapping.values():
                elm = list(filter(lambda x: self._filter_mapping[x] == elm, self._filter_mapping)).pop()
            elif predicate(elm):
                v = '?{}'.format(uuid())
                self._filter_mapping[v] = elm
                elm = v
            return elm

        s, p, o = tp_parts(tp)
        s = __create_var(s, lambda x: '<' in x and '>' in x)
        o = __create_var(o, lambda x: '"' in x or ('<' in x and '>' in x))
        return '{} {} {}'.format(s, p, o)

    def _generalize_gp(self):
        # Create a filtered graph pattern from the request one (general_gp)
        general_gp = GraphPattern()
        for new_tp in map(lambda x: self._remove_tp_filters(x), self._graph_pattern):
            general_gp.add(new_tp)
        if self._filter_mapping:
            # Store the filter mapping
            self._pipe.hmset('{}filters'.format(self._request_key), self._filter_mapping)
        return general_gp

    @abstractmethod
    def _save(self, action, general=True):
        """
        Stores data relating to the recovery of a fragment for this request
        """

        super(FragmentSink, self)._save(action)

        # Recover pattern from the request object
        self._graph_pattern = action.request.pattern

        effective_gp = self._generalize_gp() if general else self._graph_pattern

        # fragment_mapping is a tuple like (fragment_id, mapping)
        fragment_mapping = self.__check_gp_mappings(gp=effective_gp)
        exists = fragment_mapping is not None

        # Decide to proceed depending on whether it's the first time this request is received and the fragment
        # is already known
        proceed = action.id in self.passed_requests or (
            random() > 1.0 - PASS_THRESHOLD if not exists else random() > PASS_THRESHOLD)
        if not proceed:
            self.do_pass(action)
        if action.id in self.passed_requests:
            self.passed_requests.remove(action.id)

        if not exists:
            # If there is no mapping, register a new fragment collection for the general graph pattern
            fragment_id = str(uuid())
            self._fragment_key = self.__f_key_pattern.format(fragment_id)
            self._pipe.sadd(self._fragments_key, fragment_id)
            self._pipe.sadd('{}:gp'.format(self._fragment_key), *effective_gp)
            mapping = {str(k): str(k) for k in action.request.variable_labels}
            mapping.update({str(k): str(k) for k in self._filter_mapping})
        else:
            fragment_id, mapping = fragment_mapping
            self._fragment_key = self.__f_key_pattern.format(fragment_id)
            # Remove the sync state if the fragment is on-demand mode
            if r.get('{}:on_demand'.format(self._fragment_key)) is not None:
                self._pipe.delete('{}:sync'.format(self._fragment_key))

        # Here the following is persisted: mapping, pref_labels, fragment-request links and the original graph_pattern
        self._pipe.hmset('{}map'.format(self._request_key), mapping)
        if action.request.preferred_labels:
            self._pipe.sadd('{}pl'.format(self._request_key), *action.request.preferred_labels)
        self._pipe.sadd('{}:requests'.format(self._fragment_key), self._request_id)
        self._pipe.hset('{}'.format(self._request_key), 'fragment_id', fragment_id)
        self._pipe.sadd('{}gp'.format(self._request_key), *self._graph_pattern)
        self._pipe.hset('{}'.format(self._request_key), 'pattern', ' . '.join(self._graph_pattern))

        # Populate attributes that may be required during the rest of the submission process
        self._dict_fields['mapping'] = mapping
        self._dict_fields['preferred_labels'] = action.request.preferred_labels
        self._dict_fields['fragment_id'] = fragment_id

        if not exists:
            log.info('Request {} has started a new fragment collection: {}'.format(self._request_id, fragment_id))
        else:
            log.info('Request {} is going to re-use fragment {}'.format(self._request_id, fragment_id))
            n_fragment_reqs = r.scard('{}:requests'.format(self._fragment_key))
            log.info('Fragment {} is supporting {} more requests'.format(fragment_id, n_fragment_reqs))

    @abstractmethod
    def _remove(self, pipe):
        """
        Removes data relating to the recovery of a fragment for this request
        """
        fragment_id = r.hget('{}'.format(self._request_key), 'fragment_id')
        self._fragment_key = self.__f_key_pattern.format(fragment_id)
        pipe.srem('{}:requests'.format(self._fragment_key), self._request_id)
        pipe.delete('{}gp'.format(self._request_key))
        pipe.delete('{}map'.format(self._request_key))
        pipe.delete('{}pl'.format(self._request_key))
        pipe.delete('{}filters'.format(self._request_key))
        super(FragmentSink, self)._remove(pipe)

    @abstractmethod
    def _load(self):
        """
        Loads data relating to the recovery of a fragment for this request
        """
        super(FragmentSink, self)._load()
        self._graph_pattern = GraphPattern(r.smembers('{}gp'.format(self._request_key)))
        self._fragment_pattern = GraphPattern(r.smembers('{}:gp'.format(self._fragment_key)))
        self._filter_mapping = r.hgetall('{}filters'.format(self._request_key))
        self._dict_fields['mapping'] = r.hgetall('{}map'.format(self._request_key))
        self._dict_fields['preferred_labels'] = set(r.smembers('{}pl'.format(self._request_key)))

    def map(self, v, fmap=False):
        """
        Maps a fragment variable to the corresponding variable or filter of the request
        :param v: Fragment variable
        :param fmap: Map to filter or not
        :return: The mapped variable (or filter)
        """
        if self.mapping is not None:
            v = self.mapping.get(v, v)
            if fmap:
                v = self._filter_mapping.get(v, v)
        return v

    @property
    def gp(self):
        """
        :return: The request Graph Pattern
        """
        return self._graph_pattern

    @property
    def fragment_gp(self):
        """
        :return: The Graph Pattern of the supporting fragment
        """
        return self._fragment_pattern

    @property
    def filter_mapping(self):
        """
        :return: The mapping dictionary between original filters and on-the-fly created variables
        """
        return self._filter_mapping


class FragmentResponse(DeliveryResponse):
    __metaclass__ = ABCMeta

    def __init__(self, rid):
        super(FragmentResponse, self).__init__(rid)

    @abstractmethod
    def build(self):
        super(FragmentResponse, self).build()



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  This file is part of the Smart Developer Hub Project:
    http://www.smartdeveloperhub.org

  Center for Open Middleware
        http://www.centeropenmiddleware.com/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2015 Center for Open Middleware.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at 

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
import StringIO
import uuid

__author__ = 'Fernando Serena'

import pika
import sys
from rdflib import Graph, URIRef, RDF, Literal
from rdflib.namespace import Namespace, FOAF
import os
from datetime import datetime

CURATOR = Namespace('http://www.smartdeveloperhub.org/vocabulary/curator#')
TYPES = Namespace('http://www.smartdeveloperhub.org/vocabulary/types#')
AMQP = Namespace('http://www.smartdeveloperhub.org/vocabulary/amqp#')

accepted = False


def callback(ch, method, properties, body):
    if 'state' in properties.headers:
        if properties.headers['state'] == 'end':
            channel.stop_consuming()
            return
    print 'chunk headers: ', properties.headers
    for _ in eval(body):
        print _


def accept_callback(ch, method, properties, body):
    global accepted
    if not accepted:
        g = Graph()
        g.parse(StringIO.StringIO(body), format='turtle')
        if len(list(g.subjects(RDF.type, CURATOR.Accepted))) == 1:
            print 'Request accepted!'
            accepted = True


connection = pika.BlockingConnection(pika.ConnectionParameters(
    host='localhost'))
channel = connection.channel()

routing_key = ''
exchange = ''

graph = Graph()
script_dir = os.path.dirname(__file__)
with open(os.path.join(script_dir, 'query_usernames.ttl')) as f:
    graph.parse(file=f, format='turtle')


req_node = list(graph.subjects(RDF.type, CURATOR.QueryRequest)).pop()
message_id = Literal(str(uuid.uuid4()), datatype=TYPES.UUID)
agent_id = Literal(str(uuid.uuid4()), datatype=TYPES.UUID)
graph.set((req_node, CURATOR.messageId, message_id))
graph.set((req_node, CURATOR.submittedOn, Literal(datetime.now())))
agent_node = list(graph.subjects(RDF.type, FOAF.Agent)).pop()
graph.set((agent_node, CURATOR.agentId, agent_id))

ch_node = list(graph.subjects(RDF.type, CURATOR.DeliveryChannel)).pop()

result = channel.queue_declare(exclusive=True)
queue_name = result.method.queue
# channel.queue_bind(exchange=exchange, queue=queue_name, routing_key=routing_key)
channel.basic_consume(callback, queue=queue_name, no_ack=True)

result = channel.queue_declare(exclusive=True)
accept_queue = result.method.queue
channel.queue_bind(exchange='sdh', queue=accept_queue, routing_key='curator.response.{}'.format(str(agent_id)))
channel.basic_consume(accept_callback, queue=accept_queue, no_ack=True)

# graph.set((ch_node, AMQP.queueName, Literal(queue_name)))
graph.set((ch_node, AMQP.routingKey, Literal(queue_name)))
graph.set((ch_node, AMQP.exchangeName, Literal(exchange)))
message = graph.serialize(format='turtle')


channel.basic_publish(exchange='sdh',
                      routing_key='curator.request.query',
                      body=message)

channel.start_consuming()



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  This file is part of the Smart Developer Hub Project:
    http://www.smartdeveloperhub.org

  Center for Open Middleware
        http://www.centeropenmiddleware.com/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2015 Center for Open Middleware.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""

import StringIO
import calendar
import logging

from redis.lock import Lock

from abc import abstractproperty, abstractmethod, ABCMeta
from agora.stoa.actions.core import PassRequest, AGENT_ID
from agora.stoa.actions.core.utils import CGraph
from agora.stoa.store import r
from rdflib import RDF
from shortuuid import uuid

__author__ = 'Fernando Serena'

log = logging.getLogger('agora.stoa.actions.base')

log.info('Cleaning agent lock...')
r.delete('{}:lock'.format(AGENT_ID))


def agent_lock():
    lock_key = '{}:lock'.format(AGENT_ID)
    return r.lock(lock_key, lock_class=Lock)


class Action(object):
    """
    Generic action class that supports requests to Stoas
    """
    __metaclass__ = ABCMeta

    def __init__(self, message):
        """
        Base constructor
        :param message: the incoming request (RDF)
        :return:
        """
        self.__message = message
        self.__action_id = None
        self.__request_id = None

    @abstractproperty
    def request(self):
        pass

    @classmethod
    def response_class(cls):
        """
        This method should be implemented on each final Action class
        :return: The response class
        """
        pass

    @abstractproperty
    def sink(self):
        """
        :return: The sink instance
        """
        pass

    @property
    def request_id(self):
        """
        Every action is assigned a unique request id.
        """
        return self.__request_id

    @property
    def id(self):
        """
        In order to uniquely identify action requests, each incoming one produces an action id based on
        the message id and the submitter id.
        """
        return self.__action_id

    @abstractmethod
    def submit(self):
        """
        Base method that parses and saves the request message
        :return: The new request id
        """
        if not issubclass(self.response_class(), Response):
            raise SystemError(
                'The response class for this action is invalid: {}'.format(self.response_class()))
        log.info('Parsing request message...')
        self.request.parse(self.__message)
        self.__action_id = u'{}@{}'.format(self.request.message_id, self.request.submitted_by)
        self.__request_id = self.sink.save(self)
        return self.__request_id


class Sink(object):
    """
     Every action should have a sink that deals with Action persistency.
    """
    __metaclass__ = ABCMeta
    passed_requests = set([])

    def __init__(self):
        self._pipe = r.pipeline(transaction=True)
        self._request_id = None
        self._request_key = None
        self._dict_fields = {}
        self._requests_key = '{}:requests'.format(AGENT_ID)

    def load(self, rid):
        """
        Checks the given request id and loads its associated data
        """
        lock = agent_lock()
        lock.acquire()
        try:
            if not r.keys('{}:requests:{}:'.format(AGENT_ID, rid)):
                raise ValueError('Cannot load request: Unknown request id {}'.format(rid))
            self._request_id = rid
            self._request_key = '{}:requests:{}:'.format(AGENT_ID, self._request_id)
            self._load()
        finally:
            lock.release()

    @staticmethod
    def __response_fullname(f):
        """
        The response class name has to to be stored so as to be instanced on load
        :return: A string like <module>.<Response> based on the given response class function provider (f)
        """

        def wrapper():
            clz = f()
            parts = clz.__module__.split('.')
            if parts:
                module_name = parts[-1]
                return '{}.{}'.format(module_name, clz.__name__)
            raise NameError('Invalid response class: {}'.format(clz))

        return wrapper

    @abstractmethod
    def _load(self):
        """
        Sink-specific load statements (to be extended)
        :return:
        """
        self._dict_fields = r.hgetall(self._request_key)

    def __getattr__(self, item):
        if item in self._dict_fields:
            value = self._dict_fields[item]
            if value == 'True' or value == 'False':
                value = eval(value)
            return value
        return super(Sink, self).__getattribute__(item)

    def save(self, action):
        """
        Generates a new request id and stores all action data
        :return: The new request id
        """
        lock = agent_lock()
        lock.acquire()
        try:
            self._request_id = str(uuid())
            self._pipe.multi()
            self._save(action)
            # It is not until this point when the pipe is executed!
            # If it fails, nothing is stored
            self._pipe.execute()
            log.info("""Request {} was saved:
                        -message id: {}
                        -submitted on: {}
                        -submitted by: {}""".format(self._request_id, action.request.message_id,
                                                    action.request.submitted_on, action.request.submitted_by))
            return self._request_id
        finally:
            lock.release()

    def remove(self):
        """
        Creates a pipe to remove all stored data of the current request
        """
        # All dict_fields are being removed automatically here (hashmap request attributes)
        lock = agent_lock()
        lock.acquire()
        try:
            with r.pipeline(transaction=True) as p:
                p.multi()
                action_id = r.hget(self._request_key, 'id')
                p.zrem(self._requests_key, action_id)
                r_keys = r.keys('{}*'.format(self._request_key))
                for key in r_keys:
                    p.delete(key)
                self._remove(p)
                p.execute()
            log.info('Request {} was removed'.format(self._request_id))
        finally:
            lock.release()

    @abstractmethod
    def _remove(self, pipe):
        """
        Sink-specific remove statements (to be extended)
        :param pipe: The pipe to be used on remove statements
        """
        pass

    @abstractmethod
    def _save(self, action):
        """
        Sink-specific save statements (to be extended)
        :param action: The action that contains the data to be stored
        """
        # Firstly, we have to check if the action was previously stored...
        if r.zscore('{}:requests'.format(AGENT_ID), action.id):
            raise ValueError('Duplicated request: {}'.format(action.id))
        submitted_by_ts = calendar.timegm(action.request.submitted_on.timetuple())

        # The action id is stored in a sorted set using its timestamp as score
        self._pipe.zadd(self._requests_key, submitted_by_ts, action.id)
        self._request_key = '{}:requests:{}:'.format(AGENT_ID, self._request_id)
        # Basic request data is stored on a dictionary (hashmap)
        self._pipe.hmset(self._request_key, {'submitted_by': action.request.submitted_by,
                                             'submitted_on': action.request.submitted_on,
                                             'message_id': action.request.message_id,
                                             'id': self._request_id,
                                             '__response_class': self.__response_fullname(action.response_class)(),
                                             'type': action.__class__.__module__,
                                             '__hash': action.id})

    @staticmethod
    def do_pass(action):
        Sink.passed_requests.add(action.id)
        raise PassRequest()

    @property
    def request_id(self):
        return self._request_id


class Request(object):
    """
    The generic class that knows how to parse action messages and supports specific action requests
    """

    def __init__(self):
        from agora.stoa.actions.core import STOA, AMQP
        # Since the message is RDF, we create a graph to store and query its triples
        self._graph = CGraph()
        self._graph.bind('stoa', STOA)
        self._graph.bind('amqp', AMQP)
        self._request_node = None

        # Base fields dictionary
        self._fields = {}

    def parse(self, message):
        """
        Parses the message and extracts all useful content
        :param message: The request message (RDF)
        """
        log.debug('Parsing message...')
        try:
            self._graph.parse(StringIO.StringIO(message), format='turtle')
        except Exception, e:
            raise SyntaxError(e.message)

        self._extract_content()

    @abstractmethod
    def _extract_content(self, request_type=None):
        """
        Request-specific method to query the message graph and extract key content
        """
        q_res = self._graph.query("""SELECT ?node ?m ?d ?a WHERE {
                                        ?node stoa:messageId ?m;
                                              stoa:submittedOn ?d;
                                              stoa:submittedBy [
                                                 stoa:agentId ?a
                                              ]
                                     }""")
        q_res = list(q_res)
        if len(q_res) != 1:
            raise SyntaxError('Invalid request')

        request_fields = q_res.pop()

        if not all(request_fields):
            raise SyntaxError('Missing fields for generic request')

        (self._request_node, self._fields['message_id'],
         self._fields['submitted_on'],
         self._fields['submitted_by']) = request_fields

        if request_type is not None:
            request_types = set(self._graph.objects(self._request_node, RDF.type))
            if len(request_types) != 1 or request_type not in request_types:
                raise SyntaxError('Invalid request type declaration')

        log.debug(
            """Parsed attributes of generic action request:
                -message id: {}
                -submitted on: {}
                -submitted by: {}""".format(
                self._fields['message_id'], self._fields['submitted_on'], self._fields['submitted_by']))

    @property
    def message_id(self):
        return self._fields['message_id'].toPython()

    @property
    def submitted_by(self):
        return self._fields['submitted_by'].toPython()

    @property
    def submitted_on(self):
        return self._fields['submitted_on'].toPython()


class Response(object):
    """
    Every request should have a response.
    """

    __metaclass__ = ABCMeta

    def __init__(self, rid):
        self._request_id = rid

    @abstractmethod
    def build(self):
        """
        :return: A generator that provides the response
        """
        pass

    @abstractproperty
    def sink(self):
        """
        :return: The associated request sink
        """
        pass
